#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var path = require("path");
var fs_1 = require("fs");
var ts_simple_ast_1 = require("ts-simple-ast");
var ts_simple_ast_2 = require("ts-simple-ast");
var mkdirp = require("mkdirp");
var yargs_1 = require("yargs");
var ast = new ts_simple_ast_1.default();
var tab = '    ';
var mockClassFilenameSuffix = '.mock.ts';
var rootPath = process.cwd().replace(/\\/g, '/');
var providedSourcePaths;
var srcDir = rootPath + (yargs_1.argv.src ? yargs_1.argv.src.replace(/^(\.)/g, '') : '/src');
var outDir = rootPath + (yargs_1.argv.out ? yargs_1.argv.out.replace(/^(\.)/g, '') : '/test/spec');
var inputs = yargs_1.argv._;
if (inputs && inputs.length > 0) {
    providedSourcePaths = inputs.map(function (args) { return rootPath + args.replace(/^(\.)/g, ''); });
}
else {
    providedSourcePaths = [(srcDir + '/**/*.ts')];
}
ast.addExistingSourceFiles(providedSourcePaths);
ast.getSourceFiles().forEach(function (sourceFile) {
    var sourceClasses = sourceFile.getClasses();
    if (!sourceClasses.length) {
        return;
    }
    var sourcePath = sourceFile.getFilePath();
    var outputPath = sourcePath.replace(srcDir, outDir).replace(/\.ts$/, mockClassFilenameSuffix);
    var outputDirname = path.dirname(outputPath);
    var relativePath = path.relative(outputDirname, sourcePath).replace(/\\/g, '/');
    var mockedImports = sourceClasses
        .map(function (sourceClass) {
        return "import { " + sourceClass.getName() + " } from '" + relativePath.replace(/\.ts$/, '') + "';";
    });
    var mockedClasses = sourceClasses
        .map(function (sourceClass) {
        return createClassHeader(sourceClass).concat(prefix(tab, createHelpers(sourceClass)), prefix(tab, createMembers(sourceClass)), getCustomCode(sourceClass), createClassFooter(sourceClass));
    })
        .reduce(function (acc, next) {
        return acc.concat(next);
    }, []);
    var mockedSource = [
        '// tslint:disable',
        ''
    ].concat(mockedImports, [
        ''
    ], mockedClasses, [
        '',
    ]).join('\n');
    try {
        fs_1.statSync(outputDirname);
    }
    catch (err) {
        mkdirp.sync(outputDirname);
    }
    fs_1.writeFileSync(outputPath, mockedSource);
    console.info(outputPath);
});
function prefix(str, lines) {
    return lines.map(function (line) { return line !== '' ? str + line : ''; });
}
function upperCamelCase(str) {
    return str[0].toUpperCase() + str.slice(1);
}
function createTypeParams(sourceClass) {
    var params = sourceClass.getTypeParameters().map(function () { return 'any'; });
    if (!params.length) {
        return '';
    }
    return "<" + params.join(',') + ">";
}
function createClassHeader(sourceClass) {
    var className = sourceClass.getName();
    var typeParams = createTypeParams(sourceClass);
    return ["export class Mock" + className + " extends " + className + typeParams + " {"];
}
function createClassFooter(sourceClass) {
    return ["}"];
}
function createHelpers(sourceClass) {
    var className = sourceClass.getName();
    return [
        "/**",
        " * Static Helpers",
        " */",
        "",
        "private static $spies: any = {};",
        "private static get $class(): any {",
        tab + "return " + className + ";",
        "}",
        "public static $get( field: string ): any {",
        tab + "return this.$class[field];",
        "}",
        "public static $set( field: string, value: any ): any {",
        tab + "this.$class[field] = value;",
        "}",
        "public static $call( field: string, ...args: any[]): any {",
        tab + "return this.$class[field].call( this, ...args );",
        "}",
        "public static $createGetterFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOnProperty( this.$class, field, 'get' );",
        tab + "return this.$spies[field];",
        "}",
        "public static $createSetterFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOnProperty( this.$class, field, 'set' );",
        tab + "return this.$spies[field];",
        "}",
        "public static $createSpyFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOn( this.$class, field );",
        tab + "return this.$spies[field];",
        "}",
        "public static $getSpyFor( field: string ): jasmine.Spy {",
        tab + "return this.$spies[field];",
        "}",
        "",
        "/**",
        " * Instance Helpers",
        " */",
        "",
        "private $spies: any = {};",
        "private get $instance(): any {",
        tab + "return this;",
        "}",
        "private get $prototype(): any {",
        tab + "return " + className + ".prototype;",
        "}",
        "public $get( field: string ): any {",
        tab + "return this.$instance[field];",
        "}",
        "public $set( field: string, value: any ): any {",
        tab + "this.$instance[field] = value;",
        "}",
        "public $call( field: string, ...args: any[]): any {",
        tab + "return this.$prototype[field].call( this, ...args );",
        "}",
        "public $createGetterFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOnProperty( this.$instance, field, 'get' );",
        tab + "return this.$spies[field];",
        "}",
        "public $createSetterFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOnProperty( this.$instance, field, 'set' );",
        tab + "return this.$spies[field];",
        "}",
        "public $createSpyFor( field: string ): jasmine.Spy {",
        tab + "this.$spies[field] = spyOn( this.$instance, field );",
        tab + "return this.$spies[field];",
        "}",
        "public $getSpyFor( field: string ): jasmine.Spy {",
        tab + "return this.$spies[field];",
        "}",
    ];
}
function createMethod(sourceProperty) {
    var lines = [''];
    var propertyName = sourceProperty.getName();
    lines.push("/**", " * " + propertyName, " */");
    if (sourceProperty.getStaticKeyword()) {
        if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected || sourceProperty.getScope() === ts_simple_ast_2.Scope.Private) {
            lines.push("public static $call" + upperCamelCase(propertyName) + "( ...args: any[]) {", tab + "return this.$call( '" + propertyName + "', ...args );", "}");
        }
        lines.push("public static $createSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createSpyFor( '" + propertyName + "' );", "}", "public static $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    else {
        if (sourceProperty.getAbstractKeyword()) {
            lines.push("public " + propertyName + "( ...args: any[]) {", tab + "return undefined as any;", "}");
        }
        else if (sourceProperty instanceof ts_simple_ast_2.MethodDeclaration && sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public " + propertyName + "( ...args: any[]) {", tab + "return this.$call( '" + propertyName + "', ...args );", "}");
        }
        else if (sourceProperty instanceof ts_simple_ast_2.MethodDeclaration && sourceProperty.getScope() === ts_simple_ast_2.Scope.Private) {
            lines.push("public $call" + upperCamelCase(propertyName) + "( ...args: any[]) {", tab + "return this.$call( '" + propertyName + "', ...args );", "}");
        }
        lines.push("public $createSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createSpyFor( '" + propertyName + "' );", "}", "public $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    return lines;
}
function createGetter(sourceProperty) {
    var lines = [''];
    var propertyName = sourceProperty.getName();
    lines.push("/**", " * " + propertyName, " */");
    if (sourceProperty.getStaticKeyword()) {
        if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private || sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public static $get" + upperCamelCase(propertyName) + "() {", tab + "return this.$get( '" + propertyName + "' );", "}");
        }
        lines.push("public static $createGetterFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createGetterFor( '" + propertyName + "' );", "}", "public static $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    else {
        if (sourceProperty.getAbstractKeyword()) {
            lines.push("public get " + propertyName + "(): any {", tab + "return undefined as any;", "}");
        }
        else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private || sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public $get" + upperCamelCase(propertyName) + "() {", tab + "return this.$get( '" + propertyName + "' );", "}");
        }
        lines.push("public $createGetterFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createGetterFor( '" + propertyName + "' );", "}", "public $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    return lines;
}
function createSetter(sourceProperty) {
    var lines = [''];
    var propertyName = sourceProperty.getName();
    lines.push("/**", " * " + propertyName, " */");
    if (sourceProperty.getStaticKeyword()) {
        lines.push("public static $createSetterFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createSetterFor( '" + propertyName + "' );", "}", "public static $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    else {
        if (sourceProperty.getAbstractKeyword()) {
            lines.push("public set " + propertyName + "( val: any ) {", "}");
        }
        else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private || sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public $set" + upperCamelCase(propertyName) + "( val: any ) {", tab + "this.$set( '" + propertyName + "', val );", "}");
        }
        lines.push("public $createSetterFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$createSetterFor( '" + propertyName + "' );", "}", "public $getSpyFor" + upperCamelCase(propertyName) + "() {", tab + "return this.$getSpyFor( '" + propertyName + "' );", "}");
    }
    return lines;
}
function createProperty(sourceProperty) {
    var lines = [''];
    var propertyName = sourceProperty.getName();
    if (!propertyName) {
        return [];
    }
    lines.push("/**", " * " + propertyName, " */");
    if (sourceProperty.getStaticKeyword()) {
        if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public static " + propertyName + ": any;");
        }
        else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private) {
            lines.push("public static $get" + upperCamelCase(propertyName) + "() {", tab + "return this.$get( '" + propertyName + "' );", "}", "public static $set" + upperCamelCase(propertyName) + "( val: any ) {", tab + "this.$set( '" + propertyName + "', val );", "}");
        }
    }
    else {
        if (sourceProperty.getAbstractKeyword()) {
            lines.push("public " + propertyName + ": any;");
        }
        else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
            lines.push("public " + propertyName + ": any;");
        }
        else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private) {
            lines.push("public $get" + upperCamelCase(propertyName) + "() {", tab + "return this.$get( '" + propertyName + "' );", "}", "public $set" + upperCamelCase(propertyName) + "( val: any ) {", tab + "this.$set( '" + propertyName + "', val );", "}");
        }
    }
    return lines;
}
function createParameter(sourceProperty) {
    var lines = [''];
    var propertyName = sourceProperty.getName();
    if (!propertyName) {
        return [];
    }
    lines.push("/**", " * " + propertyName, " */");
    if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Protected) {
        lines.push("public " + propertyName + ": any;");
    }
    else if (sourceProperty.getScope() === ts_simple_ast_2.Scope.Private) {
        lines.push("public $get" + upperCamelCase(propertyName) + "() {", tab + "return this.$get( '" + propertyName + "' );", "}", "public $set" + upperCamelCase(propertyName) + "( val: any ) {", tab + "return this.$set( '" + propertyName + "', val );", "}");
    }
    return lines;
}
/**
 * ts-simple-ast sourceClass.getAllMembers() does not return abstract methods.
 * Github Issue: https://github.com/dsherret/ts-simple-ast/issues/102
 */
function getAbstractMethods(sourceClass) {
    var list = sourceClass.getChildSyntaxList();
    return list
        .getChildren()
        .filter(function (node) { return node.getKindName() === 'MethodDeclaration' && node.getAbstractKeyword(); });
}
function getParentClass(sourceClass) {
    var ext = sourceClass.getExtends();
    if (!ext) {
        return null;
    }
    var parentName = ext.getText().split('<')[0];
    var importSpecifiers = ext
        .getSourceFile()
        .getImportDeclarations()
        .filter(function (imp) { return imp.getNamedImports().filter(function (named) { return named.getText() === parentName; }).length; });
    if (!importSpecifiers || importSpecifiers.length == 0) {
        // TODO: the result can be empty if the parent class is in the same file.
        // This requires implementation to go through classes on current file when this happens
        return null;
    }
    var parentPath = importSpecifiers[0].getModuleSpecifier();
    var sourceFileDir = path.dirname(sourceClass.getSourceFile().getFilePath());
    var resolvedPath = path.resolve(sourceFileDir, parentPath + '.ts');
    var parentFile = ast.getSourceFile(resolvedPath);
    if (!parentFile) {
        return null;
    }
    var parentClass = parentFile.getClasses().filter(function (c) { return c.getName() === parentName; })[0];
    return parentClass;
}
function getInheritedMembers(sourceClass) {
    var parent = getParentClass(sourceClass);
    if (!parent) {
        return [];
    }
    return getAllMembers(parent);
}
function getRespectiveMockClass(sourceClass) {
    var sourceFilePath = sourceClass.getSourceFile().getFilePath();
    var mockFilePath = sourceFilePath.replace(srcDir, outDir).replace(/\.ts$/, mockClassFilenameSuffix);
    var ast = new ts_simple_ast_1.default();
    ast.addExistingSourceFiles(mockFilePath);
    var mockFile = ast.getSourceFile(mockFilePath);
    if (!mockFile) {
        return null;
    }
    var mockClass = mockFile.getClasses().filter(function (c) { return c.getName() === 'Mock' + sourceClass.getName(); })[0];
    return mockClass;
}
function getCustomCode(sourceClass) {
    var mockClass = getRespectiveMockClass(sourceClass);
    if (!mockClass) {
        return [];
    }
    var mockClassContent = mockClass.getText().split('\n');
    mockClassContent.splice(0, 1);
    mockClassContent.pop();
    var markerIndex;
    for (var index = mockClassContent.length; index > -1; index--) {
        var line = mockClassContent[index];
        if (line && line.search(/(\/\/ mockgen:custom)/g) > -1) {
            markerIndex = index;
            break;
        }
    }
    if (markerIndex) {
        return mockClassContent.slice(markerIndex, mockClassContent.length);
    }
    return [];
}
function createMembers(sourceClass) {
    var addedMembers = { instance: {}, static: {} };
    return getAllMembers(sourceClass)
        .concat(getAbstractMethods(sourceClass))
        .concat(getInheritedMembers(sourceClass))
        .map(function (sourceProperty) {
        var name = sourceProperty.getName && sourceProperty.getName();
        var isStatic = sourceProperty.getStaticKeyword && sourceProperty.getStaticKeyword();
        var added = isStatic ? addedMembers.static : addedMembers.instance;
        if (added[name]) {
            return [];
        }
        else {
            added[name] = true;
        }
        if (sourceProperty instanceof ts_simple_ast_2.MethodDeclaration) {
            return createMethod(sourceProperty);
        }
        else if (sourceProperty instanceof ts_simple_ast_2.PropertyDeclaration) {
            return createProperty(sourceProperty);
        }
        else if (sourceProperty instanceof ts_simple_ast_2.GetAccessorDeclaration) {
            return createGetter(sourceProperty);
        }
        else if (sourceProperty instanceof ts_simple_ast_2.SetAccessorDeclaration) {
            return createSetter(sourceProperty);
        }
        else if (sourceProperty instanceof ts_simple_ast_2.ParameterDeclaration) {
            return createParameter(sourceProperty);
        }
        else {
            return [];
        }
    })
        .reduce(function (acc, next) {
        return acc.concat(next);
    }, []);
}
function getParameterProperties(sourceClass) {
    if (sourceClass) {
        var properties = sourceClass.getInstanceProperties();
        return properties.filter(function (property) {
            if (property instanceof ts_simple_ast_2.ParameterDeclaration) {
                return true;
            }
        });
    }
    return [];
}
function getAllMembers(sourceClass) {
    var members = [];
    if (sourceClass) {
        members.push.apply(members, sourceClass.getMembers());
        members.push.apply(members, getParameterProperties(sourceClass));
    }
    return members;
}
