#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var path = require("path");
var fs_1 = require("fs");
var ts_simple_ast_1 = require("ts-simple-ast");
var mkdirp = require("mkdirp");
var yargs_1 = require("yargs");
var ast = new ts_simple_ast_1.default();
var tab = '    ';
var specFilenameSuffix = '.spec.ts';
var rootPath = process.cwd().replace(/\\/g, '/');
var providedSourcePaths;
var srcDir = rootPath + (yargs_1.argv.src ? yargs_1.argv.src.replace(/^(\.)/g, '') : '/src');
var outDir = rootPath + (yargs_1.argv.out ? yargs_1.argv.out.replace(/^(\.)/g, '') : '/test/spec');
var inputs = yargs_1.argv._;
if (inputs && inputs.length > 0) {
    providedSourcePaths = inputs.map(function (args) { return rootPath + args.replace(/^(\.)/g, ''); });
}
else {
    providedSourcePaths = [(srcDir + '/**/*.ts')];
}
ast.addExistingSourceFiles(providedSourcePaths);
ast.getSourceFiles().forEach(function (sourceFile) {
    var sourceClasses = sourceFile.getClasses();
    if (!sourceClasses.length)
        return;
    var sourcePath = sourceFile.getFilePath();
    var outputPath = getRespectiveSpecFilePath(sourceFile);
    if (fs_1.existsSync(outputPath)) {
        processExistingSpecFile(sourceFile);
        return;
    }
    var outputDirname = path.dirname(outputPath);
    var relativePath = path.relative(outputDirname, sourcePath).replace(/\\/g, '/');
    var specImports = sourceClasses
        .map(function (sourceClass) {
        return "import { " + sourceClass.getName() + " } from '" + relativePath.replace(/\.ts$/, '') + "';";
    });
    var generatedSpec = sourceClasses
        .map(function (sourceClass) {
        return createSpecOutline(getSpecOutlineFromSourceClass(sourceClass)).slice();
    })
        .reduce(function (acc, next) {
        return acc.concat(next);
    }, []);
    var outputSpec = specImports.concat(generatedSpec, [
        '',
    ]).join('\n');
    try {
        fs_1.statSync(outputDirname);
    }
    catch (err) {
        mkdirp.sync(outputDirname);
    }
    fs_1.writeFileSync(outputPath, outputSpec);
    console.info(outputPath);
});
function prefix(str, lines) {
    return lines.map(function (line) { return line !== '' ? str + line : ''; });
}
function createDescribeBlock(classMemberName, body) {
    if (body === void 0) { body = ['']; }
    var lines = [''];
    lines.push.apply(lines, ["describe( '" + classMemberName + "', () => {"].concat(prefix(tab, body), ["});"]));
    return lines;
}
function createSpecOutline(classSpecOutline) {
    var body = [];
    classSpecOutline.members.forEach(function (classMemberName) {
        body.push.apply(body, createDescribeBlock(classMemberName));
    });
    body.push('');
    return createDescribeBlock(classSpecOutline.className, body);
}
function getRespectiveSpecFilePath(sourceFile) {
    var sourceFilePath = sourceFile.getFilePath();
    var specFilePath = sourceFilePath.replace(srcDir, outDir).replace(/\.ts$/, specFilenameSuffix);
    return specFilePath;
}
function getRespectiveSpecFile(sourceFile) {
    var specFilePath = getRespectiveSpecFilePath(sourceFile);
    var ast = new ts_simple_ast_1.default();
    ast.addSourceFileIfExists(specFilePath);
    return ast.getSourceFile(specFilePath);
}
function getExpressionNodes(nodes) {
    return nodes.filter(function (node) { return node.getKind() == ts_simple_ast_1.SyntaxKind.ExpressionStatement; });
}
function getExpressionIdentifierName(expression) {
    if (!expression) {
        return '';
    }
    var identifierNode = expression.getChildren()[0];
    return identifierNode.compilerNode.getText();
}
function isDescribeExpression(expression) {
    return getExpressionIdentifierName(expression) === 'describe';
}
function getArgumentsFromExpression(expression) {
    var expressionArgsSyntaxList = expression.getChildren()[2];
    var expressionArgs = expressionArgsSyntaxList.getChildren();
    return expressionArgs.filter(function (argument) { return argument.getKind() != ts_simple_ast_1.SyntaxKind.CommaToken; });
}
function filterDescribeStatements(statements) {
    return statements.filter(function (statement) {
        return (statement.getKind() == ts_simple_ast_1.SyntaxKind.ExpressionStatement)
            && isDescribeExpression(statement.getExpression());
    });
}
function getSpecFileStructure(classLevelDescibeStatements) {
    var specFileStructure = {};
    classLevelDescibeStatements.forEach(function (expressionStatement) {
        var expressionArgs = getArgumentsFromExpression(expressionStatement.getExpression());
        var className = expressionArgs[0].getText().replace(/\'/g, ''); // First arg is the class name
        var testFn = expressionArgs[1]; // 2nd arg is the specification fn
        var memberStatements = filterDescribeStatements(testFn.getStatements());
        var classMemberNames = memberStatements.map(function (statement) {
            var memberArguments = getArgumentsFromExpression(statement.getExpression());
            var memberName = memberArguments[0].getText().replace(/\'/g, '');
            return memberName;
        });
        specFileStructure[className] = {
            members: classMemberNames,
            syntaxList: testFn.getChildSyntaxList()
        };
    });
    return specFileStructure;
}
function processExistingSpecFile(sourceFile) {
    var specFile = getRespectiveSpecFile(sourceFile);
    if (specFile) {
        var expectedSpecFileStructure_1 = [];
        sourceFile.getClasses().forEach(function (sourceClass) {
            var specOutline = getSpecOutlineFromSourceClass(sourceClass);
            expectedSpecFileStructure_1.push(specOutline);
        });
        var classLevelDescibeStatements = filterDescribeStatements(specFile.getStatements());
        var specFileStructure_1 = getSpecFileStructure(classLevelDescibeStatements);
        expectedSpecFileStructure_1.forEach(function (classSpecOutline) {
            var specClassBlock = specFileStructure_1[classSpecOutline.className];
            if (specClassBlock && specClassBlock.syntaxList) {
                var syntaxList_1 = specClassBlock.syntaxList;
                classSpecOutline.members.forEach(function (member) {
                    if (specClassBlock.members.indexOf(member) == -1) {
                        syntaxList_1.addChildText(createDescribeBlock(member).join('\n'));
                    }
                });
            }
            else {
                var classLevelSyntaxList = specFile && specFile.getChildSyntaxList();
                if (classLevelSyntaxList) {
                    var child = createSpecOutline(classSpecOutline).join('\n');
                    classLevelSyntaxList.addChildText(child);
                }
                else {
                    console.error('Error: This should not happen, syntax tree missing');
                }
            }
        });
        specFile.saveSync();
        console.info(specFile.getFilePath());
    }
}
function getSpecOutlineFromSourceClass(sourceClass) {
    var classMemberNames = [];
    var classMembers = sourceClass.getMembers();
    classMembers.forEach(function (classMember) {
        if (classMember instanceof ts_simple_ast_1.MethodDeclaration
            || classMember instanceof ts_simple_ast_1.GetAccessorDeclaration
            || classMember instanceof ts_simple_ast_1.SetAccessorDeclaration) {
            if (classMember.getScope() === ts_simple_ast_1.Scope.Protected
                || classMember.getScope() === ts_simple_ast_1.Scope.Public) {
                classMemberNames.push(classMember.getName());
            }
        }
    });
    var classSpecOutline = {
        className: sourceClass.getName(),
        members: classMemberNames
    };
    return classSpecOutline;
}
